@@ -10,8 +10,14 @@ fn main() {
     }
 
     let before_writes = std::time::Instant::now();
-    for i in 0_u64..1_000_000_000 {
+    for i in 1_u64..1_000_000_000 {
         lsm.insert(i.to_le_bytes(), [0; 8]).unwrap();
+        if i % 1_000_000 == 0 {
+            println!(
+                "{:.2} million wps",
+                i as f64 / (before_writes.elapsed().as_micros() + 1) as f64
+            )
+        }
     }
     lsm.flush().unwrap();
     dbg!(before_writes.elapsed());
